structure Ast = struct

    type symbol = Symbol.symbol
    type integer = LargeInt.int

    type pos = int
    type region = pos * pos

    datatype typ =
	NAMEDtyp of symbol
      | INTtyp
      | BOOLtyp
      | ARRtyp of typ
      | MARKtyp of region * typ

    datatype binop =
	LESSop
      | GREATERop
      | LESSEQop
      | GREATEREQop
      | EQUALop
      | NOTEQUALop
      | PLUSop
      | MINUSop
      | TIMESop

    datatype uop =
	NOTop
      | UMINUSop

    datatype lvalue =
	IDlval of symbol
      | SUBlval of exp * exp
      | SELlval of exp * symbol

    and exp =
	CONDexp of exp * exp * exp
      |	BINOPexp of binop * exp * exp
      | UOPexp of uop * exp
      | CALLexp of exp * exp list
      | INTexp of integer
      | BOOLexp of bool
      | THISexp
      | NULLexp
      | NEWOBJexp of symbol
      | NEWARRexp of typ * exp
      | FETCHexp of lvalue
      | ASSIGNexp of lvalue * exp
      | MARKexp of region * exp

    datatype init =
	EXPinit of exp
      | ARRinit of init list

    type locvardecl = { name: symbol, typ: typ, init: init option,
			region: region }

    datatype forinit =
	EXPlist of exp list
      | LOCVARdecls of locvardecl list

    datatype stm =
	IGNstm of exp
      | SEQstm of stm * stm
      | IFstm of exp * stm * stm
      | WHILEstm of exp * stm
      | DOstm of stm * exp
      | FORstm of forinit * exp * exp list * stm
      | RETURNstm of exp option
      | NULLstm
      | LABELEDstm of symbol * stm
      | BREAKstm of symbol option
      | CONTINUEstm of symbol option
      | LETstm of locvardecl list * stm
      | MARKstm of region * stm

    type methoddecl = { public: bool,
			restyp: typ option,
			name: symbol,
			formals: (symbol * typ) list,
			body: stm,
			region: region }

    type instvardecl = { public: bool, typ: typ, name: symbol, region: region }

    type classdecl = { name : symbol,
		       superclass : symbol option,
		       vardecls : instvardecl list,
		       methoddecls : methoddecl list,
		       region: region }

    type mainclass = { classname: symbol,
		       methodname: symbol,
		       formal: symbol * typ,
		       body: stm,
		       region: region}

    type program = { main: mainclass, classes: classdecl list }


    fun binopName LESSop = "<"
      | binopName GREATERop =  ">"
      | binopName LESSEQop = "<="
      | binopName GREATEREQop = ">="
      | binopName EQUALop = "=="
      | binopName NOTEQUALop = "!="
      | binopName PLUSop = "+"
      | binopName MINUSop = "-"
      | binopName TIMESop = "*"

    fun uopName NOTop = "!"
      | uopName UMINUSop = "-"
end
