structure Translate :> sig

    type exp
    type lexp

    type frame

    type frag

    val dummy : exp
    val ldummy : lexp

    val cond : exp * exp * exp -> exp
    val compare : Ast.binop * exp * exp -> exp
    val equal : Ast.binop * exp * exp -> exp
    val arith : Ast.binop * exp * exp -> exp
    val not : exp -> exp
    val neg : exp -> exp
    val int : LargeInt.int -> exp
    val bool : bool -> exp
    val nilexp : exp
    val newobj : Types.typ -> exp
    val newarr : exp -> exp
    val fetch : lexp -> exp
    val assign : lexp * exp -> exp
    val call :  exp * int * exp list -> exp

    val newtemp : unit -> lexp
    val sublval : exp * exp -> lexp
    val sellval : exp * int -> lexp
    val arrlength : exp -> lexp

    val whilestm : { clab: Label.label,
		     blab: Label.label, c: exp, b: exp } -> exp
    val dostm : { clab: Label.label, blab: Label.label, b: exp, c: exp } -> exp
    val forstm : { clab: Label.label, blab: Label.label,
		   init: exp list, c: exp, rinit: exp list, b: exp } -> exp
    val discard : exp -> exp
    val seq : exp list -> exp
    val ifstm : exp * exp * exp -> exp
    val jump : Label.label -> exp
    val label : Label.label -> exp
    val nullstm: exp

    val codefrag : frame * exp -> frag
    val classfrag : Types.classrec -> frag

    val frame : int * Types.typ option -> frame
    val main : unit -> frame
    val resinfo : frame -> (Types.typ * lexp) option
    val thisacc : frame -> lexp
    val retlab : frame -> Label.label
    val argaccl : frame -> lexp list
    val entrylab : frame -> Label.label

end = struct

    type lexp = unit
    type exp = unit
    type frame = { resinfo : (Types.typ * lexp) option,
		   retlab: Label.label, entrylab: Label.label,
		   argaccl : lexp list }
    type frag = unit

    val dummy = () : exp
    val ldummy = () : lexp

    fun efun _ = dummy
    fun lfun _ = ldummy
    fun ffun _ = () : frag

    val cond = efun
    val compare = efun
    val equal = efun
    val arith = efun
    val not = efun
    val neg = efun
    val int = efun
    val bool = efun
    val nilexp = dummy
    val newobj = efun
    val newarr = efun
    val fetch = efun
    val assign = efun
    val call = efun

    val newtemp = lfun
    val sublval = lfun
    val sellval = lfun
    val arrlength = lfun

    val whilestm = efun
    val dostm = efun
    val forstm = efun
    val discard = efun
    val seq = efun
    val ifstm = efun
    val jump = efun
    val label = efun
    val nullstm = dummy

    val codefrag = ffun
    val classfrag = ffun

    fun frame (i, ri) = { resinfo = Option.map (fn ti => (ti, ldummy)) ri,
			  retlab = Label.new (),
			  entrylab = Label.new (),
			  argaccl = List.tabulate (i, lfun) }
    fun main () = frame (0, NONE)
    fun resinfo (f: frame) = #resinfo f
    val thisacc = lfun
    fun retlab (f: frame) = #retlab f
    fun entrylab (f: frame) = #entrylab f
    fun argaccl (f: frame) = #argaccl f
end
