structure Parse : sig
    val parse : string -> Ast.program
end =
struct 
  structure MiniJavaLrVals =
      MiniJavaLrValsFun(structure Token = LrParser.Token)
  structure Lex = MiniJavaLexFun(structure Tokens = MiniJavaLrVals.Tokens)
  structure MiniJavaP = JoinWithArg(structure ParserData = MiniJavaLrVals.ParserData
				    structure Lex=Lex
				    structure LrParser = LrParser)

  val errcons = ErrorMsg.defaultConsumer ()

  fun parse filename =
      let val file = TextIO.openIn filename
	  val source = Source.newSource (filename, 1, file, false, errcons)
	  val sm = #sourceMap source
	  fun error r m = ErrorMsg.error source r ErrorMsg.COMPLAIN m ErrorMsg.nullErrorBody
	  val depth = ref 0
	  fun enterC () = depth := !depth + 1
	  fun leaveC () = let val d = !depth - 1 in depth := d; d = 0 end
	  fun newline pos = SourceMap.newline sm pos
	  fun handleEof () = let
	      val pos = SourceMap.lastChange sm
	  in
	      if !depth > 0 then
		  error (pos, pos) "unexpected end of input in comment"
	      else ();
	      Source.closeSource source;
	      pos
	  end
	  fun get _ = TextIO.input file
	  val lexarg =
	      { enterC = enterC,
		leaveC = leaveC,
		handleEof = handleEof,
		newline = newline,
		error = error }
	  val lexer = MiniJavaP.makeLexer get lexarg
	  val (ast, _) = MiniJavaP.parse(30,lexer,
					 fn (s, p, p') => error (p, p') s,
					 ())
       in
	  Semant.semant (ast, source);
	  ast
      end handle LrParser.ParseError => raise ErrorMsg.Error

end
