structure Types = struct

    structure SlotMap = IntRedBlackMap
    structure NameMap = Symbol.Map

    datatype typ =
        NULLtyp
      | VOIDtyp
      | INTtyp
      | BOOLtyp
      | ARRtyp of typ
      | CLASStyp of Symbol.symbol * class ref
      | ERRORtyp

    and class =
	CLASS of classrec
      | ERRORcls

    withtype method = { formals: typ list,
			restyp: typ option,
			slot: int,	    (* for dynamic access *)
			label: Label.label, (* for creating the class fragment *)
			public: bool }

    and instvar = { name: Symbol.symbol,
		    slot: int,
		    typ: typ,
		    public: bool }

    and classrec = { stamp: int,
		     depth: int,
		     name: Symbol.symbol,
		     superclasses: class list,
		     instvars: instvar SlotMap.map,
		     instvarenv: instvar NameMap.map,
		     methods: method NameMap.map }

    fun toString INTtyp = "int"
      | toString BOOLtyp = "boolean"
      | toString NULLtyp = "null"
      | toString VOIDtyp = "void"
      | toString (ARRtyp t) = toString t ^ "[]"
      | toString (CLASStyp (sy, _)) = Symbol.toString sy
      | toString ERRORtyp = "?"
end
